# -*- coding: future_fstrings -*-
# This module  contains (eventually all) functions that relate to the 
# logging process
import copy
import numpy as np
import os
import pickle
import psutil as psu
import sys
import time
import traceback
import warnings

from datetime import datetime
from inspect import stack
from pyFAT_astro.Support.fat_errors import ProgramError,InputError

with warnings.catch_warnings():
   warnings.simplefilter("ignore")
   import matplotlib
   matplotlib.use('pdf')
   import matplotlib.pyplot as plt
   import matplotlib.font_manager as mpl_fm


class full_system_tracking:
   def __init__(self,Configuration):
      self.stop = False
      self.pid = os.getpid()
      self.main_pyFAT = psu.Process(self.pid)
      self.user = self.main_pyFAT.username()
      self.python = self.main_pyFAT.name()
      self.tirific = Configuration['TIRIFIC']
      self.font_file = Configuration['FONT_FILE']
      self.sofia = Configuration['SOFIA2']
      self.file = f"{Configuration['MAIN_DIRECTORY']}FAT_Resources_Used.txt"
      self.plot_name= f"{Configuration['MAIN_DIRECTORY']}pyFAT_Resources_Monitor.pdf"
      self.cpus= psu.cpu_count()
      with open(self.file,'w') as resources:
         resources.write("# This file contains an estimate of all resources used for a pyFAT run. \n")
         resources.write(f"# {'Time':20s} {'Sys CPU':>10s} {'Sys RAM':>10s} {'FAT CPU':>10s} {'FAT RAM':>10s} \n")
         resources.write(f"# {'YYYY-MM-DD hh:mm:ss':20s} {'%':>10s} {'Gb':>10s} {'%':>10s} {'Gb':>10s} \n")
      if not 'ALL' in Configuration['DEBUG_FUNCTION']:
         self.interval = 5
      else:
         self.interval = 30 # amount of second when to do new monitor

   def start_monitoring(self):
      while not self.stop:
         try:
               self.sys_cpu= psu.cpu_percent(interval=1)
               self.sys_ram= psu.virtual_memory().used/2**30.
               self.CPU = 0.
               self.RAM = 0.
               for proc in psu.process_iter():
                  if proc.username() == self.user \
                     and proc.status() == 'running'\
                     and (proc.name() == self.python or\
                        proc.name() == self.tirific or\
                        proc.name() == self.sofia or\
                        proc.name() == 'python3'):
                     try:
                           self.CPU += proc.cpu_percent(interval=0.5)/self.cpus
                           self.RAM += (proc.memory_info()[0])/2**30.
                     except:
                           pass
               #file.write(f"{datetime.now()} CPU = {CPU} % Mem = {mem} Gb for TiRiFiC \n")
               with open(self.file,'a') as resources:
                  resources.write(f"{datetime.now().strftime('%Y-%m-%d %H:%M:%S'):20s} {self.sys_cpu:>10.1f} {self.sys_ram:>10.2f} {self.CPU:>10.1f} {self.RAM:>10.2f} \n")
         except Exception as e:
               #We do not care if something goes wrong once. We don't want the monitor to crash
               #but we would like to know what went wrong
               traceback.print_exception(type(e),e,e.__traceback__)
               pass
         time.sleep(self.interval)

   def stop_monitoring(self):
      self.interval = 0.1
      self.stop = True
      try:
         loads = {'Time':[]}
         keys=['SCPU','SRAM','FCPU','FRAM']
         for key in keys:
               loads[key]  = []
         with open(self.file) as file:
               lines = file.readlines()
         startdate = 0
         #load data from file into dictionary
         for line in lines:
               line = line.split()
               if line[0] == '#':
                  continue
               else:
                  date = extract_date(f"{line[0]} {line[1]}")
               if startdate == 0:
                  startdate = date
               diff = date - startdate
               time = diff.total_seconds()/(3600.)
               loads['Time'].append(time)
               for i,key in enumerate(keys):
                  loads[key].append(float(line[2+i]))
         #Plot the parameters
         try:
               mpl_fm.fontManager.addfont(self.font_file)
               font_name = mpl_fm.FontProperties(fname=self.font_file).get_name()
         except FileNotFoundError:
               font_name = 'DejaVu Sans'
         
         labelfont = {'family': font_name,
                  'weight': 'normal',
                  'size': 4}
         fig, ax1 = plt.subplots(figsize = (8,6))
         fig.subplots_adjust(left = 0.1, right = 0.9, bottom = 0.3, top = 0.7)
         ax1.plot(loads['Time'],loads['SRAM'],'b--',lw=0.5,alpha=0.5, label='System RAM')
         ax1.plot(loads['Time'],loads['FRAM'],'b-',lw=0.5,alpha=1.0, label='pyFAT RAM')
         ax1.set_ylim(0,np.max(np.array(loads['SRAM']+loads['FRAM'],dtype=float))*1.1)
         ax1.set_ylabel('RAM (Gb) ', color='b')
         ax1.tick_params(axis='y', labelcolor='b')
         ax1.set_xlabel('Run Duration (h)', color='k',zorder=5)
         ax2 = ax1.twinx()
         ax2.plot(loads['Time'],loads['SCPU'],'r--',lw=0.5,alpha=0.5, label='System CPU')
         ax2.plot(loads['Time'],loads['FCPU'],'r-',lw=0.5,alpha=1.0, label='pyFAT CPU')
         ax2.set_ylim(0,np.max(np.array(loads['SCPU']+loads['FCPU'],dtype=float))*1.1)
         ax2.set_ylabel('CPUs (%)',color='r')
         ax2.tick_params(axis='y', labelcolor='r')
         fig.legend(bbox_to_anchor=(0.9, 0.705),loc='lower right')
         fig.savefig(self.plot_name)
         plt.close()
      except Exception as e:
         traceback.print_exception(type(e),e,e.__traceback__)
         print(f'We failed to write the statistics plot.')
         pass
full_system_tracking.__doc__ =f'''
NAME:
   full_system_tracking

PURPOSE:
   A class that can be started in a thread and run in the background to track 
   all instances of tirific sofia 2 and the main python id   

 CATEGORY:
    log_functions 

 INPUTS:
   Configuration = Standard FAT configuration
   

 OPTIONAL INPUTS:


 OUTPUTS:


 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
 
'''


def enter_recovery_point(Configuration, Fits_Files = None, Tirific_Template= None,\
                         IP= None,
                         message = 'You should have specified a message',point_ID='RANDOM'):
   update_statistics(Configuration, message= message)
   print_log(f'''ENTER_RECOVERY_POINT: Creating a recovery point.
{message}              
''',Configuration,case= ['debug_start','main'])
   Configuration['RP_SECTION'] = point_ID
   name = copy.deepcopy(point_ID)
   if Configuration['RP_SECTION'] == 'ITERATION':
      Configuration['RP_COUNTER'] += 1
      name = f'{name}_{Configuration["RP_COUNTER"]}'
   write_config(f"{Configuration['LOG_DIRECTORY']}RP_{name}.pkl",\
      Configuration, Fits_Files=Fits_Files, IP = IP,\
                 Tirific_Template=Tirific_Template,message=message )
enter_recovery_point.__doc__ =f'''
 NAME:
   enter_recovery_point
 PURPOSE:
   use pickle to enter a recovery point where the current configuration and other 
   tracking is written.

 CATEGORY:
    log_functions 

 INPUTS:
   Configuration = Standard FAT configuration
   Fits_Files = standard FAT Fits Files
   Tirific_Template = 
   message = message for the point
   point_ID = Name of Recovery point

 OPTIONAL INPUTS:


 OUTPUTS:


 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
 
'''

def extract_date(string):
    tmp = string.split(' ')
    tmp2 = tmp[0].split('-')
    if len(tmp2) == 3:
        try:
            date =  datetime.strptime(f"{tmp[0]} {tmp[1]}", '%Y-%m-%d %H:%M:%S.%f')
        except ValueError:
            date =  datetime.strptime(f"{tmp[0]} {tmp[1]}", '%Y-%m-%d %H:%M:%S')
    else:
        raise ProgramError("There is no date in the provided string.")
    return date
extract_date.__doc__ =f'''
 NAME:
    extract_date

 PURPOSE:
    convert a string into a date object

 CATEGORY:
    write_functions

 INPUTS:
    string = string

 OPTIONAL INPUTS:


 OUTPUTS:
    date = the date object

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''
    

def get_usage_statistics(Configuration,process):
        # psutil returns bytes
    memory_in_mb = (process.memory_info()[0])/2**20. 
    cpu_percent = process.cpu_percent(interval=1)
    return cpu_percent,memory_in_mb

get_usage_statistics.__doc__ =f'''
 NAME:
    get_usage_statistics
 PURPOSE:
    use psutil to get the current CPU and memory usage of tirific

 CATEGORY:
    log_functions 

 INPUTS:
    Configuration = Standard FAT configuration
    process_id = process id of the tirific currently running.

 OPTIONAL INPUTS:


 OUTPUTS:
    CPU = The current CPU usage
    mem = current memory usage in Mb

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
    pyFAT version < 1.0.0 uses top which only works on unix and has an 
    error in the MB calculation.
'''

def return_help_message():
   help_message = '''
   Use pyFAT in this way for batch fitting:

      pyFAT configuration_file=FAT_Input.yml

   where configuration_file specifies a yaml file with specific settings
   such as the catalog.

   For fitting use pyFAT in this way:

      pyFAT cube_name=Input_Cube.fits

   Where Input_Cube.fits is the observation to be fitted. In this mode
   configuration_file can still be used to specify fit settings but
   catalogue and location setting will be ignored.

      pyFAT -h

   prints this message.

      pyFAT print_examples=True

   prints a yaml file (FAT_defaults.yml)  with the default values for all
   possible fitting parameters and an example input catalogue (FAT_Example_Catalogue.txt).
   The files are printed in the current working directory. In the yaml
   file values designated ??? indicated values without defaults.

   All config parameters can be set directly from the command e.g:

      pyFAT cube_name=Input_Cube.fits fitting.ring_size=1.5 'fitting.fixed_parameters=[INCL,SDIS]'

   You can test your installation with:

      pyFAT installation_check=True

   '''
   return help_message

# A simple function to return the line numbers in the stack from where
# the functions are called.
def linenumber(debug='short'):
    '''get the line number of the print statement in the main.'''
    line = []
    for key in stack():
        if key[1] == 'main.py':
            break
        if key[3] != 'linenumber' and key[3] != 'print_log' and key[3] != '<module>':
            file = key[1].split('/')
            to_add= f"In the function {key[3]} at line {key[2]}"
            if debug == 'long':
                to_add = f"{to_add} in file {file[-1]}."
            else:
                to_add = f"{to_add}."
            line.append(to_add)
    if len(line) > 0:
        if debug == 'long':
            line = ', '.join(line)+f'\n{"":8s}'
        elif debug == 'short':
            line = line[0]+f'\n{"":8s}'
        else:
            line = f'{"":8s}'
    else:
        for key in stack():
            if key[1] == 'main.py':
                line = f"{'('+str(key[2])+')':8s}"
                break
    return line

linenumber.__doc__ =f'''
 NAME:
    linenumber

 PURPOSE:
    get the line number of the print statement in the main. Not sure 
    how well this is currently working.

 CATEGORY:
    log_functions 

 INPUTS:

 OPTIONAL INPUTS:


 OUTPUTS:
    the line number of the print statement

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
    If debug = True the full stack of the line print will be given, 
    in principle the first debug message in every function should set 
    this to true and later messages not.
    !!!!Not sure whether currently the linenumber is produced due to 
    the restructuring.
'''

def print_log(log_statement,Configuration, case = None):
   if case is None:
      case=['main']
   debugging = False
   debug= 'empty'
   if Configuration['DEBUG']:
      trig=False
      if 'ALL' in Configuration['DEBUG_FUNCTION']:
         trig = True
      else:
         # get the function  
         for key in stack():
            if key[3] != 'linenumber' and key[3] != 'print_log' and key[3] != '<module>': 
               current_function= f"{key[3]}"
               break
         if current_function.lower() in [x.lower() for x in  Configuration['DEBUG_FUNCTION']]:
            trig=True      
      if trig:
         debugging=True    
         if 'debug_start' in case:
            debug = 'long'
         else:
            debug= 'short'
   if Configuration['TIMING']:
      log_statement = f"{linenumber(debug=debug)} {datetime.now()} {log_statement}"
   else:
      log_statement = f"{linenumber(debug=debug)}{log_statement}"
   print_statement = False
 
   if (debugging and ('debug_start' in case or 'debug_add' in case))\
      or ('verbose' in case and (Configuration['VERBOSE_LOG'] or debugging))\
         or 'main' in case:
            print_statement = True
   
   if print_statement:
      if Configuration['VERBOSE_SCREEN'] \
         or Configuration['OUTPUTLOG'] is None  \
         or 'screen' in case \
         or (Configuration['VERBOSE_LOG'] and 'main' in case):
         print(log_statement)
      if Configuration['OUTPUTLOG']:
         with open(Configuration['OUTPUTLOG'],'a') as log_file:
            log_file.write(log_statement)

print_log.__doc__ =f'''
 NAME:
    print_log
 PURPOSE:
    Print statements to log if existent and screen if Requested
 CATEGORY:
    log_functions 

 INPUTS:
    log_statement = statement to be printed
    Configuration = Standard FAT Configuration

 OPTIONAL INPUTS:


    screen = False
    also print the statement to the screen

 OUTPUTS:
    line in the log or on the screen

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    linenumber, .write

 NOTE:
    If the log is None messages are printed to the screen.
    This is useful for testing functions.
'''

def update_statistics(Configuration,process= None,message = None ):
    if Configuration['TIMING']:
        function = traceback.format_stack()[-2].split('\n')
        function = function[0].split()[-1].strip()
        if process is None:
            process = Configuration['FAT_PSUPROCESS']
            program = 'pyFAT'
        else:
            program = 'TiRiFiC'

        CPU,mem = get_usage_statistics(Configuration,process)
        with open(f"{Configuration['LOG_DIRECTORY']}Usage_Statistics.txt",'a') as file:
            if message:
                file.write(f"# {function.upper()}: {message} at {datetime.now()} \n")
            # We cannot copy the process so initialize in the configuration
            file.write(f"{datetime.now()} CPU = {CPU} % Mem = {mem} Mb for {program} \n")
update_statistics.__doc__ =f'''
 NAME:
    update_statistics
 PURPOSE:
    Update the files that track the usage statistics 
 CATEGORY:
    log_functions

 INPUTS:
    Configuration = Standard FAT Configuration

 OPTIONAL INPUTS:
    process = the process ID to track, if not sset the FAT_PSUPROCESS 
              in the Configuration will tracked.
    
    message = Message that will be printed at line before the usages 
              statistics are printed
    

 OUTPUTS:
    update statistic file   
 
 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
   
 NOTE:
    If the message is not provided  the CPU and Memory will be dumped 
    without an indication of the process that has been tracked.
'''


def write_config(file,Configuration, Fits_Files = None, \
      Tirific_Template = None, IP=None,message=None ):
   print_log(f'''WRITE_CONFIG: writing the configuration to {file}
''',Configuration,case=['debug_start'])
   # Separate the keyword names
   #Proper dictionaries are not pickable
   Pick_Configuration = {}
   for key in Configuration:
      Pick_Configuration[key] = Configuration[key]
      if key == 'FAT_PSUPROCESS':
         Pick_Configuration[key] = None
   if not Fits_Files is None:
      Pick_Configuration['Fits_Files'] = Fits_Files
   if not Tirific_Template is None:
      Pick_Configuration['Tirific_Template'] = Tirific_Template
   if not message is None:
      Pick_Configuration['Message'] = message
   if not IP is None:
      Pick_Configuration['Initial_Parameters'] = IP
   with open(file,'wb') as tmp:
      pickle.dump(Pick_Configuration,tmp) 
write_config.__doc__ =f'''
 NAME:
    write_config

 PURPOSE:
    Write a config file to the fitting directory.

 CATEGORY:
    write_functions

 INPUTS:
    file = name of the file to write to
    Configuration = Standard FAT configuration

 OPTIONAL INPUTS:


 OUTPUTS:
    A FAT config file.

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE: This doesn't work in python 3.6
This Could be used to pickup the fitting from where we left of

'''

def recover_recovery_point_conf(Configuration_In):
   #first check that the file exists
   conf = f"{Configuration_In['MAIN_DIRECTORY']}/{Configuration_In['LOG_DIRECTORY']}/RP_{Configuration_In['RECOVERY_POINT']}.pkl"
   if os.path.isfile(conf):
      with open(conf,'rb') as file:
         Configuration = pickle.load(file)
      if Configuration['MULTIPROCESSING']:
          print(f'We cannot recover and do multiprocessing, turning multiprocessing off. ')
          Configuration['MULTIPROCESSING'] = False
      return Configuration
   else:
      raise InputError(f'''We failed to locate the file {conf}
Please give a legit recovery point (Note that the default log is by date.)''')
     
        


# -*- coding: future_fstrings -*-

# This is the python version of FAT
import numpy as np
import os
import psutil
import pyFAT_astro
import pyFAT_astro.Support.support_functions as sf
import pyFAT_astro.Support.read_functions as rf
import sys
import traceback
import warnings
import threading


from datetime import datetime
from multiprocessing import Pool,get_context,Lock,Manager

from pyFAT_astro.FAT_Galaxy_Loop import FAT_Galaxy_Loop,MP_initialize_sofia,\
                                        MP_Fitting_Loop
from pyFAT_astro.config.defaults import process_input
from pyFAT_astro.Support.fat_errors import ProgramError,BadCatalogueError,InputError
from pyFAT_astro.Support.write_functions import reorder_output_catalogue
from pyFAT_astro.Support.log_functions import full_system_tracking

def warn_with_traceback(message, category, filename, lineno, file=None, line=None):
    log = file if hasattr(file,'write') else sys.stderr
    traceback.print_stack(file=log)
    log.write(warnings.formatwarning(message, category, filename, lineno, line))



# String syntax ''' '''for multiline strings. " " for string without break and ' ' for indexing dictionaries

#from memory_profiler import profile
#@profile
def main_trace():
    from viztracer import VizTracer
    with VizTracer(output_file="FAT_Run_Viztracer.json",min_duration=1000) as tracer:
        main()

def main():
    argv = sys.argv[1:]
    try:
        #Get default settings
        print(f"This is version {pyFAT_astro.__version__} of pyFAT.")
        if pyFAT_astro.__branch__:
            print(f"This is a github distribution and we are on the branch {pyFAT_astro.__branch__}.")
        cfg =process_input(argv)
        
        # if we set more cpus than available we limit to the available cpus
        try:
            if cfg.ncpu > len(psutil.Process().cpu_affinity()):
                cfg.ncpu  = len(psutil.Process().cpu_affinity())
        except AttributeError:
            if cfg.ncpu > psutil.cpu_count():
                cfg.ncpu  = psutil.cpu_count()

    

        #Transform all to a Configuration dictionary
        Original_Configuration = sf.setup_configuration(cfg)

        if cfg.output.debug:
            warnings.showwarning = warn_with_traceback
        if Original_Configuration['RP_SECTION'] == 'START':
            #First we check for sofia and TiRiFiC
            Original_Configuration['SOFIA2'] = \
                sf.find_program(Original_Configuration['SOFIA2'], "SoFiA 2")
            Original_Configuration['TIRIFIC'] =\
                sf.find_program(Original_Configuration['TIRIFIC'], "TiRiFiC")
            if not cfg.cube_name is None:
                Full_Catalogue = sf.Proper_Dictionary({})
                Full_Catalogue['ENTRIES'] =\
                    ['ENTRIES','ID','DISTANCE','DIRECTORYNAME','CUBENAME']
                Full_Catalogue['ID'] = [f"{cfg.cube_name.split('/')[-1]}"]
                Full_Catalogue['DISTANCE'] = [-1.]
                Full_Catalogue['DIRECTORYNAME'] = ['./']
                Full_Catalogue['CUBENAME'] = [cfg.cube_name.split('/')[-1]]
            elif 'sofia_catalogue' in Original_Configuration['FITTING_STAGES']:
                Full_Catalogue = rf.sofia_input_catalogue(Original_Configuration)
            elif not Original_Configuration['CATALOGUE'] is None:
                Full_Catalogue = rf.catalogue(Original_Configuration['CATALOGUE']\
                    ,split_char= cfg.advanced.catalogue_split_character)
            else:
                raise InputError(f'''                             
{'':8s}We could not find any of the following input: 
{'':8s}cube_name= (We found {cfg.cube_name})
{'':8s}input.catalogue= (We found { Original_Configuration['CATALOGUE']})
{'':8s}and 'sofia_catalogue' was not in the fitting stages (We found { Original_Configuration['FITTING_STAGES']})
{'':8s}Please add any of these when calling pyFAT or at them in the correct manner to the yml configuration file.
''')                        
        
            # Get the longest directory name to format the output directory properlyFit_Tirific_OSC
            for directory in Full_Catalogue['DIRECTORYNAME']:
                if directory == './':
                    directory = Original_Configuration['MAIN_DIRECTORY'].split('/')[-2]
                if len(directory) > Original_Configuration['MAXIMUM_DIRECTORY_LENGTH']:
                    Original_Configuration['MAXIMUM_DIRECTORY_LENGTH'] = len(directory)

            # Create a file to write the results to if if required
            if Original_Configuration['OUTPUT_CATALOGUE']:
                if not os.path.exists(Original_Configuration['OUTPUT_CATALOGUE']) or Original_Configuration['NEW']:
                    if os.path.exists(Original_Configuration['OUTPUT_CATALOGUE']) and Original_Configuration['NEW']:
                        os.rename(Original_Configuration['OUTPUT_CATALOGUE'],f"{os.path.splitext(Original_Configuration['OUTPUT_CATALOGUE'])[0]}_Prev.txt")
                    with open(Original_Configuration['OUTPUT_CATALOGUE'],'w') as output_catalogue:
                        comment = 'Comments on Fit Result'
                        AC1 = 'OS'
                        output_catalogue.write(f"{'Directory Name':<{Original_Configuration['MAXIMUM_DIRECTORY_LENGTH']}s} {AC1:>6s} {comment}\n")

          

              
            #if start_galaxy not negative then it is catalogue ID
            if Original_Configuration['CATALOGUE_START_ID'] in ['-1','-1.']:
                Original_Configuration['CATALOGUE_START_ID'] = int(0)
            else:
                Original_Configuration['CATALOGUE_START_ID'] = int(np.where(Original_Configuration['CATALOGUE_START_ID'] == np.array(Full_Catalogue['ID'],dtype=str))[0][0])
            # If the end galaxy is -1 fit the whole catalogue
            if Original_Configuration['CATALOGUE_END_ID'] in ['-1','-1.']:
                Original_Configuration['CATALOGUE_END_ID'] = int(len(Full_Catalogue['ID']))
                if Original_Configuration['CATALOGUE_END_ID'] == 0:
                    Original_Configuration['CATALOGUE_END_ID'] = 1
            else:
                Original_Configuration['CATALOGUE_END_ID'] = int(np.where(Original_Configuration['CATALOGUE_END_ID'] == np.array(Full_Catalogue['ID'],dtype=str))[0][0])+1
            # start the main fitting loop
            if float(Original_Configuration['CATALOGUE_START_ID']) > float(Original_Configuration['CATALOGUE_END_ID']):
                raise BadCatalogueError(f''' Your starting galaxy (Line nr = {Original_Configuration['CATALOGUE_START_ID']}) is listed after your ending galaxy (Line nr = {Original_Configuration['CATALOGUE_END_ID']}), maybe you have double catalogue ids?''')
       
        if Original_Configuration['TIMING']:
            if Original_Configuration['RP_SECTION'] == 'START':
                with open(Original_Configuration['MAIN_DIRECTORY']+'Timing_Result.txt','w') as timing_result:
                    timing_result.write("Timing results for every section of the fit process for all galaxies.  \n")
                # If we do this we should have 1 cpu to keep going
                Original_Configuration['NCPU'] -= 1
            system_monitor = full_system_tracking(Original_Configuration)
            fst = threading.Thread(target=system_monitor.start_monitoring)
            fst.start()
            print(f"We are using {Original_Configuration['NCPU']} cpus for fitting and 1 for timing.")
        else:
            print(f"We are using {Original_Configuration['NCPU']} cpus.")
           
        
        #if we do recovery multiprocessing is turned off but the points should still be written       
        if Original_Configuration['MULTIPROCESSING']:
            Original_Configuration['VERBOSE_SCREEN'] = False
            #output_catalogue = copy.deepcopy(Original_Configuration['OUTPUT_CATALOGUE'])
            #Original_Configuration['OUTPUT_CATALOGUE'] = None
            no_processes,sofia_processes = sf.calculate_number_processes(Original_Configuration)
            Configs_and_Locks = []

            with Manager() as loop_manager:
                timing_lock = loop_manager.Lock()
                catalogue_lock = loop_manager.Lock()
                #In case of multiprocessing we want to make sure to start with
                #The big galaxies
                #Setup an array of configs with locks
                for current_galaxy_index in range(Original_Configuration['CATALOGUE_START_ID'], Original_Configuration['CATALOGUE_END_ID']):
                    Configs_and_Locks.append([sf.set_individual_configuration(current_galaxy_index,Full_Catalogue,Original_Configuration),timing_lock,catalogue_lock])

                #Get all intitial setups
                with get_context("spawn").Pool(processes=sofia_processes) as pool:
                    print(f'Starting size estimates with {sofia_processes} processes')
                    initial_setups = pool.starmap(MP_initialize_sofia, Configs_and_Locks)

                initial_setups = [x for x in initial_setups if x['Succes']]
                sizes = np.array([np.mean(x['Size']) for x in initial_setups]\
                    ,dtype=float)
                if len(sizes) > 0.:
                    sorted_ind = np.flip(sizes.argsort())
                    sorted_initial_setups = [[initial_setups[x],timing_lock,catalogue_lock] \
                        for x in sorted_ind]
                    initial_setups =[]
                    with get_context("spawn").Pool(processes=no_processes) as pool:
                        print(f'Starting fitting with {no_processes} processes')
                        finals = pool.starmap(MP_Fitting_Loop, sorted_initial_setups)

                else:
                    print(f'All galaxies can not be fitted')

            #For clarity we reorder the output results to match the input
            reorder_output_catalogue(Original_Configuration,Full_Catalogue)
            #Stitch all temporary outpu catalogues back together
            #with open(output_catalogue,'a') as catalogue:
            #    for x in results:
            #        catalogue.writelines(x)

        else:
            # as in recovery we might be switching from MP to single reset PER_GALAXY_NCPU
            Original_Configuration['PER_GALAXY_NCPU'] = sf.set_limits(Original_Configuration['NCPU'],1,20)
            for current_galaxy_index in range(Original_Configuration['CATALOGUE_START_ID'], Original_Configuration['CATALOGUE_END_ID']):
                if Original_Configuration['RP_SECTION'] == 'START':
                    Configuration = sf.set_individual_configuration(current_galaxy_index,Full_Catalogue,Original_Configuration)
                else:
                    Configuration = Original_Configuration
                catalogue_line = FAT_Galaxy_Loop(Configuration)
        
        if Original_Configuration['TIMING']:
            system_monitor.stop_monitoring()
            fst.join()

    except SystemExit:
        try:
            system_monitor.stop_monitoring()
            fst.join()
        except:
            pass
        pass
    except KeyboardInterrupt:
        traceback.print_exception(*sys.exc_info())
        try:
            system_monitor.stop_monitoring()
            fst.join()
        except:
            pass
        pass
    except InputError:
        traceback.print_exception(*sys.exc_info())
        try:
            system_monitor.stop_monitoring()
            fst.join()
        except:
            pass
        pass
    except:
        try:
            system_monitor.stop_monitoring()
            fst.join()
        except:
            pass
        raise ProgramError(f'''Something went wrong in the main. This should not happen. Please list an issue on github.''')

main.__doc__ = '''
 NAME:
     main
 PURPOSE:
      Fit Tilted Ring Models with Tirific in a fully automated manner
 CATEGORY:
      Main for fitting galaxies. Tirific still requires interactive fitting this code attempts
      to remedy that

 CALLING SEQUENCE:
     see pyFAT -h

 INPUTS:
    see pyFAT -h

 OUTPUTS:
     See Readme or just run the code

 EXAMPLE:
     pyFAT  configuration_file=/home/your_computer/FAT_dir/FAT_INPUT.yml'
'''



# -*- coding: future_fstrings -*-



import os
from dataclasses import dataclass, field
from datetime import datetime
from omegaconf import OmegaConf,MISSING
#from multiprocessing import cpu_count
import psutil
import sys
from typing import List, Optional

try:
    from importlib.resources import files as import_pack_files
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    # For Py<3.9 files is not available
    from importlib_resources import files as import_pack_files
    
@dataclass
class Fitting:

    # Catalogue ID of the first galaxy to be fitted. -1 Means start at the beginning
    catalogue_start_id: str = '-1'
    # the last galaxy to be fitted, if set to -1 the whole catalogue will be fitted
    catalogue_end_id: str = '-1'
    fitting_stages: List = field(default_factory=lambda: [
                                 'Create_FAT_Cube', 'Run_Sofia', 'Fit_Tirific_OSC'])
    # Possible stages are
    # Create_FAT_Cube: Create a FAT compatible cube from the original cube. This will overwrite any previous FAT cube present. If omitted it is assumed the Cube is present in the fitting directory
    # Sofia_Catalogue: The catalogue is assumed to be a sofia catalogue and all sources in the catalogue are to be fitted
    # Run_Sofia: Run Sofia on the FAT cube and  process the output
    # Existing_Sofia: It is assumed the Sofia output exist and specified in the fitting catalogue, this means a catalogue exists for every cubelet
    # Fit_Tirific_OSC: Run FAT using the Tirific program in multiple iterations and smooth
    # Tirshaker: Bootstrap errors for the final model. This can take a long time
    ring_size: float = 1.1  # The size of the rings in number of beams
    fixed_parameters:  List = field(default_factory=lambda: [
                                    'Z0', 'XPOS', 'YPOS', 'VSYS'])  # Options are INCL, PA, SDIS, SBR
    opt_pixel_beam: int = 4

    distance: float = -1.  # Distance to the galaxy, set from the catalogue at start of loop in case of batch fitting


@dataclass
class Input:
    main_directory: str = f'{os.getcwd()}'
    # 'Options are independent, sinusoidal, hanning
    channel_dependency: str = 'independent'
    catalogue: Optional[str] = None
    tirific: str = "tirific"  # Command to call tirific
    sofia2: str = "sofia"  # Command to call sofia 2
    sofia_basename: Optional[str] = None
    # Directory of the existing Sofia output. Only used if the input catalogue is sofia or pre-processed sofia is somewhere
    sofia_dir: Optional[str] = None


@dataclass
class Output:
    # Name of the log dir
    log_directory: str = f'Logs/{datetime.now().strftime("%d-%m-%Y")}'
    log_file: str = 'Log.txt'  # Name of the log file
    catalogue: str = 'pyFAT_results.txt'
    new: bool = True  # Create all output anew
    # How much output you would like to maintain for each galaxy. 0 just organize the output and keep all (This will also happen when a fit is unsuccesful, this can be a lot of files); 1 remove optimized files, log files and input files; 2  remove optimized files, log files, input files, ps files and unsmoothed files; 3 (Default) remove optimized files, log files, input files, ps files, unsmoothed files and all model fits files except the final model; 4 keep only the def files and remove all other output. 5 indicates a failed fit clean up. >6 is the same as 0. Residuals are created for all cases where the fits files are maintained.
    quantity: int = 3
    # If you want FAT to output a warp radius, tiltograms and warp radius set warp_output (Default = n)
    warp_output: bool = False
    # The font to be used in the plots
    font_file: str = "/usr/share/fonts/truetype/msttcorefonts/Times_New_Roman.ttf"
    debug: bool = False
    timing: bool = False
    verbose_log: bool = False
    verbose_screen: bool = False

@dataclass
class Advanced:
    # The maximum number of iterations that FAT tries to calls trific bfeore it call it quits
    max_iterations: int = 15
    loops: int = 7  # The number of full loops set for tirific in a  single iteration, this is increased every time the maximum is reach upto max 15 loops
    # if the number of beams across the major axis/2. is less than this size we will only fit a flat disc,set here.
    minimum_warp_size: float = 3.
    # we need at least this amount of rings (Including 0 and 1/5 beam), set here
    minimum_rings: int = 3
    too_small_galaxy: float = 1.  # if the number of beams across the major axis/2 is less than this we will not fit the galaxy, set here
    unreliable_size: float = 2.  # If the final diameter is smaller than this the fit is considered unreliable
    unreliable_inclination: float = 10.  # If the final inclination is below this the fit is considered unreliable
    shaker_iterations: int = 20
    multiprocessing: bool = True
    number_of_disks: int = 2
    # The whole mechanism beheind the sbr_limits (which weigh the smoothing) is emperically determined. They are multiplied with this factor to allow optimization
    limit_modifier_factor: float = 1.05
    #We do not want to use too many cores per galaxy.
    per_galaxy_ncpu: int = 4
    catalogue_split_character: str = '|'
    # Allow for the user to set the boundaries in the fitting
    pa_input_boundary:  List = field(
        default_factory=lambda: [[0., 0.], [0., 0.], [0., 0.]])
    incl_input_boundary:  List = field(
        default_factory=lambda: [[0., 0.], [0., 0.], [0., 0.]])
    sdis_input_boundary:  List = field(
        default_factory=lambda: [[0., 0.], [0., 0.], [0., 0.]])
    z0_input_boundary:  List = field(
        default_factory=lambda: [[0., 0.], [0., 0.], [0., 0.]])
    vsys_input_boundary:  List = field(
        default_factory=lambda: [[0., 0.], [0., 0.], [0., 0.]])
    xpos_input_boundary:  List = field(
        default_factory=lambda: [[0., 0.], [0., 0.], [0., 0.]])
    ypos_input_boundary:  List = field(
        default_factory=lambda: [[0., 0.], [0., 0.], [0., 0.]])
    vrot_input_boundary: List = field(
        default_factory=lambda: [[0., 0.], [0., 0.], [0., 0.]])
    sbr_input_boundary: List = field(
        default_factory=lambda: [[0., 0.], [0., 0.], [0., 0.]])
    # The brightest pixels need to have a SNR above this value
    source_max_snr: float = 2.5
    # The fraction of pixels in the source that need to be above max_snr
    source_max_fraction: float = 0.075
    # The mean SNR required in the source
    source_mean_snr: float = 0.5
    sofia_threshold: int = 5
    #option to bypass all initial checks on the intial sofia source and simply force the TRM fitting
    force_fit: bool = False
    # Add the channel dependency, minimum inclination,
    debug_function: List = field(default_factory=lambda: ['ALL'] )
    # If we are using a sofia catalogue and want tor overwrite the create FAT Cubes
    sofia_overwrite: bool = False
    #We want to be able to start at a given point in the code.
    recovery_point: str = 'START_0'

@dataclass
class defaults:
    try:
        ncpu: int = len(psutil.Process().cpu_affinity())
    except AttributeError:
        ncpu: int = psutil.cpu_count()

    print_examples: bool = False
    installation_check: bool = False
    cube_name: Optional[str] = None
    configuration_file: Optional[str] = None
    input: Input = field(default_factory = Input)
    output: Output = field(default_factory = Output)
    fitting: Fitting = field(default_factory = Fitting)
    advanced: Advanced = field(default_factory = Advanced)

def process_input(argv):
    if '-v' in argv or '--version' in argv:
        #The version is always printed so only need to exit()
        sys.exit()

    if '-h' in argv or '--help' in argv:
        from pyFAT_astro.Support.log_functions import return_help_message
        print(return_help_message())
        sys.exit()
    #First import the defaults
    cfg = OmegaConf.structured(defaults)
    #if ncpu is the total available remove 1
    if cfg.ncpu == psutil.cpu_count():
        cfg.ncpu -= 1

    # read command line arguments anything list input should be set in brackets '' e.g. pyROTMOD 'rotmass.MD=[1.4,True,True]'
    inputconf = OmegaConf.from_cli(argv)
    cfg_input = OmegaConf.merge(cfg,inputconf)
    # Print examples if requested
    if cfg_input.print_examples:
        no_cube = OmegaConf.masked_copy(cfg, ['ncpu','input','output',\
            'fitting','advanced'])
        with open('FAT_defaults.yml','w') as default_write:
            default_write.write(OmegaConf.to_yaml(no_cube))
        my_resources = import_pack_files('pyFAT_astro.config')
        data = (my_resources / 'FAT_Input_Catalogue.txt').read_bytes()
        with open('FAT_Example_Catalogue.txt','w+b') as default_write:
            default_write.write(data)
        print(f'''We have printed the file FAT_defaults.yml FAT_Input_Catalogue.txt in {os.getcwd()}.
''')
        sys.exit()
    #if a configuration file is provided read it
    if not cfg_input.configuration_file is None:
        succes = False
        while not succes:
            try:
                yaml_config = OmegaConf.load(cfg_input.configuration_file)
        #merge yml file with defaults
                cfg = OmegaConf.merge(cfg,yaml_config)
                succes = True
            except FileNotFoundError:
                cfg_input.configuration_file = input(f'''
You have provided a config file ({cfg_input.configuration_file}) but it can't be found.
If you want to provide a config file please give the correct name.
Else press CTRL-C to abort.
configuration_file = ''')



    cfg = OmegaConf.merge(cfg,inputconf)
    
    if not any([cfg.cube_name, cfg.configuration_file, cfg.installation_check\
                ,cfg.print_examples,cfg.input.catalogue]):
        from pyFAT_astro.Support.fat_errors import InputError
        raise InputError(f'''You have not provided a cube_name, configuration_file or input_catalogue
nor requested the installation_check or print_examples.
We are crashes because that means you want to do nothing.''')
    
    return cfg


# -*- coding: future_fstrings -*-
# This module contains a set of functions and classes that are used in several different Python scripts in the Database.
from astropy.io import fits
from astropy.wcs import WCS
from datetime import datetime
from shutil import copyfile
from pyFAT_astro.Support.fat_errors import FunctionCallError,BadCubeError\
    ,BadMaskError,InputError
from pyFAT_astro.Support.log_functions import print_log
import pyFAT_astro.Support.support_functions as sf


from make_moments.functions import moments

#from pyFAT_astro.Support.read_functions import obtain_border_pix
from scipy import ndimage
import numpy as np
import copy
import warnings
import os

#Check that the mask only contains the selected sources
def check_mask(Configuration,id,Fits_Files,SNR= 5.):
    print_log(f'''CHECK_MASK: Checking the mask to contain only the correct source.
and to have decent signal noise
''',Configuration, case =['debug_start','verbose'])
    if Configuration['DEBUG']:
        copyfile(f"{Configuration['FITTING_DIR']}/{Fits_Files['MASK']}",\
            f"{Configuration['FITTING_DIR']}/Sofia_Output/{Configuration['BASE_NAME']}_Original_Mask.fits")
    mask = fits.open(f"{Configuration['FITTING_DIR']}/{Fits_Files['MASK']}"\
        ,uint = False, do_not_scale_image_data=True,ignore_blank = True,\
        output_verify= 'ignore')
    mask = check_mask_sources(Configuration,id,mask)

    if SNR < 3.:
        mask = smooth_mask(Configuration,mask)
    fits.writeto(f"{Configuration['FITTING_DIR']}{Fits_Files['MASK']}",mask[0].data,\
            mask[0].header, overwrite = True)
    # to ensure compatible units and calculations with th models we make the maps ourselves
    del mask[0].header['C*3']
    chan_map = np.array(np.nansum(mask[0].data,axis=0),dtype =float)
    fits.writeto(f"{Configuration['FITTING_DIR']}/Sofia_Output/{Configuration['BASE_NAME']}_chan.fits",\
        chan_map,header=mask[0].header,overwrite = True)
    mask.close()
    if Configuration['SOFIA_RAN']:
        print_log('CHECK_MASK: Creating Sofia Moments',Configuration,case=['verbose'])
        messages = moments(filename =  f"{Configuration['FITTING_DIR']}{Fits_Files['FITTING_CUBE']}",\
                        mask =  f"{Configuration['FITTING_DIR']}/{Fits_Files['MASK']}",\
                        overwrite = True, map_velocity_unit= 'km/s',\
                        debug = Configuration['DEBUG'], log=True,\
                        output_directory =   f"{Configuration['FITTING_DIR']}/Sofia_Output",\
                        output_name = f"{Configuration['BASE_NAME']}")
        print_log(messages,Configuration,case=['verbose'])
        
    
    #moments()
        #make_moments(Configuration, Fits_Files,fit_type='Generic_Initialize',vel_unit = 'm/s')

check_mask.__doc__ =f'''
 NAME:
    check_mask

 PURPOSE:
    Check that the mask only contains the source we want and if that source
    has low SNR that it is smoothed with a beam

 CATEGORY:
    fits_functions

 INPUTS:
    Configuration = Standard FAT configuration
    id = the SoFiA ID of the source we are interested in
    Fits_Files = Standard FAT dictionary with filenames
    Tirific_Template = Standard FAT Tirific Template

 OPTIONAL INPUTS:
    fit_type = 'Undefined'

 OUTPUTS:
    moment maps and the channel map

 OPTIONAL OUTPUTS:
    a cleaned mask

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''

def check_mask_sources(Configuration,id,mask):
    if float(id) not in mask[0].data:
        print_log(f'''CHECK_MASK: We cannot find the selected source in the mask. This will lead to errors. Aborting the fit.
    ''',Configuration,case=['main','screen'])
        BadMaskError(f" We can not find the sofia source id in the mask.")
    else:
        data = copy.deepcopy(mask[0].data)
        data[data != float(id)] = 0.
        diff = data-mask[0].data
        neg_index = np.where(diff < 0.)[0]
        if len(neg_index) != 0:
            print_log(f'''CHECK_MASK: The initial mask had more than a single source. redoing the mask.
    ''',Configuration, case = ['verbose'])
            mask[0].data = data


    mask[0].data[mask[0].data> 0.5] = 1

    mask[0].header['BITPIX'] = 16
    return mask
check_mask_sources.__doc__ =f'''
 NAME:
    check_mask

 PURPOSE:
    Check that the mask only contains the source we want and not parts of other nearby sources.
    Subsequently make the channel_map and the moment maps.

 CATEGORY:
    fits_functions

 INPUTS:
    Configuration = Standard FAT configuration
    id = the SoFiA ID of the source we are interested in
    Fits_Files = Standard FAT dictionary with filenames
    Tirific_Template = Standard FAT Tirific Template

 OPTIONAL INPUTS:
    fit_type = 'Undefined'

 OUTPUTS:
    moment maps and the channel map

 OPTIONAL OUTPUTS:
    a cleaned mask

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''
# Create a cube suitable for FAT
def create_fat_cube(Configuration, Fits_Files = None,sofia_catalogue=False,\
        id='No default',name='No default'):
    Configuration['PREPARATION_TIME'][0] = datetime.now()
    #First get the cubes
    if sofia_catalogue:
        Cube = fits.open(f"{Configuration['SOFIA_DIR']}{name}_{id}_cube.fits",\
            uint = False, do_not_scale_image_data=True,ignore_blank = True, output_verify= 'ignore')
    else:
        Cube = fits.open(Configuration['FITTING_DIR']+Fits_Files['ORIGINAL_CUBE'],\
            uint = False, do_not_scale_image_data=True,ignore_blank = True, output_verify= 'ignore')
    data = Cube[0].data
    hdr = Cube[0].header
    if hdr['NAXIS'] == 4:
        data = data[0,:,:,:]
        del hdr['NAXIS4']
        hdr['NAXIS'] = 3
    # clean the header
    hdr = sf.clean_header(Configuration,hdr)
    data = prep_cube(Configuration,hdr,data)
    # and write our new cube

  
    if sofia_catalogue:
        print_log(f'''CREATE_FAT_CUBE: We are writing a FAT modified cube to be used for the fitting. This cube is called {name}_{id}/{name}_{id}_FAT.fits
''',Configuration)
        try:
            fits.writeto(f"{Configuration['MAIN_DIRECTORY']}{name}_FAT_cubelets/{name}_{id}/{name}_{id}_FAT.fits",data,hdr,overwrite=Configuration['SOFIA_OVERWRITE'])
        except OSError:
            raise InputError('pyFAT has already setup this directory, if you want to overwrite the bcubelets please set advanced.sofia_overwrite to True in the configuration')

    else:
        print_log(f'''CREATE_FAT_CUBE: We are writing a FAT modified cube to be used for the fitting. This cube is called {Configuration['FITTING_DIR']+Fits_Files['FITTING_CUBE']}
''',Configuration)
        fits.writeto(Configuration['FITTING_DIR']+Fits_Files['FITTING_CUBE'],data,hdr)
    # Release the arrays

    Cube.close()

    data = []
    if sofia_catalogue:
        return hdr
    hdr = []
    Configuration['PREPARATION_TIME'][1] = datetime.now()

create_fat_cube.__doc__ =f'''
 NAME:
    create_fat_cube

 PURPOSE:
    As we do not want to work from the original cube
    this function copies that cube into a new cube
    that we can then happily modify and work with.
    To keep disk space overhead this cube might later on also
    be cut down to size.

 CATEGORY:
    fits_functions

 INPUTS:
    Configuration = Standard FAT configuration
    Fits_Files = Standard FAT dictionary with filenames

 OPTIONAL INPUTS:

 OUTPUTS:
    The FAT input cube

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''

def cut_cubes(Configuration, Fits_Files, galaxy_box):
    print_log(f'''CUT_CUBES: Starting to cut the cube_size.
''', Configuration, case= ['debug_start'])
    cube_edge= [6.,5.*round(Configuration['BEAM_IN_PIXELS'][0]),5.*round(Configuration['BEAM_IN_PIXELS'][0])]
    cube_size= []
    new_cube = []
    for i in [2,1,0]:
        cube_size.append(Configuration['NAXES'][i])
        new_cube.append([0,Configuration['NAXES'][i]])
    new_cube = np.array(new_cube,dtype=int)
    cut  = False
    for i,limit in enumerate(galaxy_box):
        if limit[0] > cube_edge[i]:
            cut = True
            new_cube[i,0] = limit[0]-sf.float_to_int(cube_edge[i])
        if limit[1] < cube_size[i] - cube_edge[i]:
            cut = True
            new_cube[i,1] = limit[1]+sf.float_to_int(cube_edge[i])
    if 'restart_fitting' in [x.lower() for x in Configuration['FITTING_STAGES']]:
        if 'create_fat_cube' in [x.lower() for x in Configuration['FITTING_STAGES']]:
            files_to_cut = [Fits_Files['FITTING_CUBE']]
        else:
            cut = False
    else:
        files_to_cut = [Fits_Files['FITTING_CUBE'],Fits_Files['MASK'],\
                        Fits_Files['MOMENT0'],\
                        Fits_Files['MOMENT1'],\
                        Fits_Files['MOMENT2'],\
                        Fits_Files['CHANNEL_MAP'],\
                        ]
    if cut:
        print_log(f'''CUT_CUBES: Your input cube is significantly larger than the detected source.
{"":8s}CUT_CUBES: we will cut to x-axis = [{new_cube[2,0]},{new_cube[2,1]}] y-axis = [{new_cube[1,0]},{new_cube[1,1]}]
{"":8s}CUT_CUBES: z-axis = [{new_cube[2,0]},{new_cube[2,1]}].
{"":8s}CUT_CUBES: We will cut the following files:
{"":8s}CUT_CUBES: {', '.join(files_to_cut)}
''', Configuration)

        for file in files_to_cut:
            print_log(f'''CUT_CUBES: We are cutting the file {file}
''', Configuration)
            if os.path.exists(f"{Configuration['FITTING_DIR']}{file}"):
                cutout_cube(Configuration,file,new_cube)
            else:
                if file == Fits_Files['CHANNEL_MAP']:
                    pass
                else:
                    raise FunctionCallError(f'We are trying to cut {file} but it does not exist')


    #We want to check if the cube has a decent number of pixels per beam.
    if not os.path.exists(f"{Configuration['FITTING_DIR']}/{Fits_Files['OPTIMIZED_CUBE']}"):
        optimized_cube(Configuration, Fits_Files)
    return new_cube
cut_cubes.__doc__ =f'''
 NAME:
    cut_cubes

 PURPOSE:
    Cut all FAT related output back in size to a system that fits snugly around the sofia detection

 CATEGORY:
    fits_functions

 INPUTS:
    Configuration = Standard FAT configuration
    Fits_Files = Standard FAT dictionary with filenames
    galaxy_box = array that contains the new size as
                 [[z_min,z_max],[y_min,y_max], [x_min,x_max]]
                 adhering to fits' idiotic way of reading fits files

 OPTIONAL INPUTS:

 OUTPUTS:

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    cutout_cube, optimized_cube

 NOTE: The cut cubes will overwrite the _FAT products
       down to size and overwrite the existing files,
       User supplied SoFIA products are hence copied as fat products.
       To keep them safe.
'''

def cutout_cube(Configuration,filename,sub_cube):
    Cube = fits.open(Configuration['FITTING_DIR']+filename,uint = False, do_not_scale_image_data=True,ignore_blank = True, output_verify= 'ignore')
    hdr = Cube[0].header

    if hdr['NAXIS'] == 3:
        data = Cube[0].data[sub_cube[0,0]:sub_cube[0,1],sub_cube[1,0]:sub_cube[1,1],sub_cube[2,0]:sub_cube[2,1]]
        hdr['NAXIS1'] = sub_cube[2,1]-sub_cube[2,0]
        hdr['NAXIS2'] = sub_cube[1,1]-sub_cube[1,0]
        hdr['NAXIS3'] = sub_cube[0,1]-sub_cube[0,0]
        hdr['CRPIX1'] = hdr['CRPIX1'] -sub_cube[2,0]
        hdr['CRPIX2'] = hdr['CRPIX2'] -sub_cube[1,0]
        hdr['CRPIX3'] = hdr['CRPIX3'] -sub_cube[0,0]
        #Only update when cutting the cube
        Configuration['NAXES'] =[ hdr['NAXIS1'],hdr['NAXIS2'],hdr['NAXIS3']]
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            coordinate_frame = WCS(hdr)
            xlow,ylow,zlow = coordinate_frame.wcs_pix2world(1,1,1., 1.)
            xhigh,yhigh,zhigh = coordinate_frame.wcs_pix2world(*Configuration['NAXES'], 1.)
            xlim = np.sort([xlow,xhigh])
            ylim = np.sort([ylow,yhigh])
            zlim =np.sort([zlow,zhigh])/1000.
            sf.set_boundaries(Configuration,'VSYS',*zlim,input=True)
            sf.set_boundaries(Configuration,'XPOS',*xlim,input=True)
            sf.set_boundaries(Configuration,'YPOS',*ylim,input=True)

    elif hdr['NAXIS'] == 2:
        data = Cube[0].data[sub_cube[1,0]:sub_cube[1,1],sub_cube[2,0]:sub_cube[2,1]]
        hdr['NAXIS1'] = sub_cube[2,1]-sub_cube[2,0]
        hdr['NAXIS2'] = sub_cube[1,1]-sub_cube[1,0]
        hdr['CRPIX1'] = hdr['CRPIX1'] -sub_cube[2,0]
        hdr['CRPIX2'] = hdr['CRPIX2'] -sub_cube[1,0]

    Cube.close()
    fits.writeto(Configuration['FITTING_DIR']+filename,data,hdr,overwrite = True)
cutout_cube.__doc__ =f'''
 NAME:
    cutout_cube

 PURPOSE:
    Cut filename back to the size of subcube, update the header and write back to disk.

 CATEGORY:
    fits_functions

 INPUTS:
    Configuration = Standard FAT configuration
    filename = name of the cube to be cut
    sub_cube = array that contains the new size as
                [[z_min,z_max],[y_min,y_max], [x_min,x_max]]
                adhering to fits' idiotic way of reading fits files.

 OPTIONAL INPUTS:

 OUTPUTS:
    the cut cube is written to disk.

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''


def prep_cube(Configuration,hdr,data):
    print_log( f'''PREPROCESSING: starting the preprocessing of the cube
''',Configuration, case = ['debug_start'])

    if hdr['CDELT3'] < -1:
        print_log( f'''PREPROCESSING: Your velocity axis is declining with increasing channels
{"":8s}PREPROCESSING: We reversed the velocity axis.
''', Configuration)
        hdr['CDELT3'] = abs(hdr['CDELT3'])
        hdr['CRPIX3'] = hdr['NAXIS3']-hdr['CRPIX3']+1
        data = data[::-1,:,:]
    #Check for zeros
    if np.where(data == 0.):
        print_log(f'''PREPROCESSING: Your cube contains values exactly 0. If this is padding these should be blanks.
{"":8s}PREPROCESSING: We have changed them to blanks.
''', Configuration)
        data[np.where(data == 0.)] = float('NaN')

    # check for blank channels and noise statistics
    cube_ok = False
    prev_first_comparison = 0.
    prev_last_comparison = 0.
    times_cut_first_channel = 0
    times_cut_last_channel = 0
    corner_box_size = int(np.floor(2*hdr['BMAJ']/abs(hdr['CDELT1'])))
    if corner_box_size > np.mean([hdr['NAXIS1'],hdr['NAXIS2']])/10.:
        corner_box_size=int(np.floor(np.mean([hdr['NAXIS1'],hdr['NAXIS2']])/10.))
    while not cube_ok:
        while np.isnan(data[0,:,:]).all():
            data=data[1:,:,:]
            hdr['NAXIS3'] = hdr['NAXIS3']-1
            hdr['CRPIX3'] = hdr['CRPIX3'] - 1
            if hdr['NAXIS3'] < 5:
                print_log(f'''PREPROCESSING: This cube has too many blanked channels.
''', Configuration,case = ['main', 'screen'])
                raise BadCubeError("The cube has too many blanked channels")
            log_statement = f'''PREPROCESSING: We are cutting the cube as the first channel is completely blank.
'''
            print_log(f'''PREPROCESSING: We are cutting the cube as the first channel is completely blank.
''', Configuration)
        while np.isnan(data[-1,:,:]).all():
            data=data[:-1,:,:]
            hdr['NAXIS3'] = hdr['NAXIS3']-1
            if hdr['NAXIS3'] < 5:
                print_log(f'''PREPROCESSING: This cube has too many blanked channels.
''', Configuration,case = ['main', 'screen'])
                raise BadCubeError("The cube has too many blanked channels")
            print_log(f'''PREPROCESSING: We are cutting the cube as the last channel is completely blank.
''', Configuration)
        #Then check the noise statistics
        noise_first_channel = np.nanstd(data[0,:,:])
        noise_last_channel = np.nanstd(data[-1,:,:])
        noise_bottom_right =  np.nanstd(data[:,:corner_box_size,-corner_box_size:])
        noise_top_right =  np.nanstd(data[:,-corner_box_size:,-corner_box_size:])
        noise_bottom_left =  np.nanstd(data[:,:corner_box_size,:corner_box_size])
        noise_top_left =  np.nanstd(data[:,-corner_box_size:,:corner_box_size])
        noise_corner = np.nanmean([noise_top_right,noise_bottom_right,noise_bottom_left,noise_top_left])
        channel_noise = np.nanmean([noise_first_channel,noise_last_channel])
        if ~np.isfinite(noise_corner):
            noise_corner = copy.deepcopy(channel_noise)
        difference = abs((noise_first_channel-noise_last_channel)/noise_first_channel)
        if noise_corner/channel_noise >1.5 and difference < 0.2:
            noise_corner = copy.deepcopy(channel_noise)
        difference2 = abs(channel_noise-noise_corner)/channel_noise
        if difference < 0.2 and np.isfinite(difference) and difference2 < 0.25:
            cube_ok = True
        else:
            print_log(f'''PREPROCESSING: We are cutting the cube as clearly the noise statistics are off.
{"":8s}PREPROCESSING: Noise in the first channel is {noise_first_channel}
{"":8s}PREPROCESSING: Noise in the last channel is {noise_last_channel}
{"":8s}PREPROCESSING: Noise in the corners is {noise_corner}
''',Configuration,case= ['verbose'])
            first_comparison = abs(noise_first_channel-noise_corner)/noise_corner
            last_comparison = abs(noise_last_channel-noise_corner)/noise_corner
            if prev_first_comparison == 0:
                prev_first_comparison = first_comparison
            if prev_last_comparison == 0.:
                prev_last_comparison = last_comparison
            if (first_comparison > last_comparison and \
               abs((first_comparison-prev_first_comparison)/first_comparison) >= abs((last_comparison-prev_last_comparison)/last_comparison) and \
               times_cut_first_channel < 0.1*hdr['NAXIS3']) or \
               ~np.isfinite(noise_first_channel):
                prev_first_comparison = first_comparison
                times_cut_first_channel += 1
                data=data[1:,:,:]
                hdr['CRPIX3'] = hdr['CRPIX3'] - 1
                hdr['NAXIS3'] = hdr['NAXIS3'] - 1
            else:
                if times_cut_last_channel < 0.1*hdr['CRPIX3'] or \
                    ~np.isfinite(noise_last_channel):
                    prev_last_comparison = last_comparison
                    times_cut_last_channel += 1
                    data=data[:-1,:,:]
                    hdr['NAXIS3'] = hdr['NAXIS3'] - 1
                else:
                    cube_ok = True

            if times_cut_first_channel >= 8 and times_cut_last_channel >= 8:
                cube_ok = True
                print_log( f'''PREPROCESSING: This cube has non-uniform noise statistics.'
''',Configuration)
            if hdr['NAXIS3'] < 5:
                print_log(f'''PREPROCESSING: This cube has noise statistics that cannot be dealt with.
''',Configuration,case = ['main','screen'])
                raise BadCubeError('The Cube has noise statistics that cannot be dealt with')
    if ~np.isfinite(noise_corner):
        print_log(f'''PREPROCESSING: This cube has noise statistics that cannot be dealt with.
''',Configuration,case = ['main','screen'])
        raise BadCubeError('The Cube has noise statistics that cannot be dealt with')
    #hdr['FATNOISE'] = noise_corner
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        low_noise_indices = np.where(data < -10*noise_corner)
    if len(low_noise_indices) > 0.0001*hdr['NAXIS1']*hdr['NAXIS2']*hdr['NAXIS3']:
        data[low_noise_indices] = float('NaN')
        print_log(f'''PREPROCESSING: Your cube had a significant amount of values below -10*sigma. If you do not have a central absorption source there is something seriously wrong with the cube.
{"":8s}PREPROCESSING: We blanked these values.
''',Configuration)

    #Let's make sure that a central absorption source is treated the same regardless od the size of the cube
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        buffer=5.*hdr['BMAJ']/np.mean([abs(hdr['CDELT1']),abs(hdr['CDELT2'])])
        central = data[:,sf.float_to_int(hdr['NAXIS2']/2.-buffer):sf.float_to_int(hdr['NAXIS2']/2.+buffer),\
                        sf.float_to_int(hdr['NAXIS1']/2.-buffer):sf.float_to_int(hdr['NAXIS1']/2.+buffer)]

        central_noise_indices = np.where(central < -10*noise_corner)
    if len(central_noise_indices) > 0.0001*4*buffer*hdr['NAXIS3']:
        central[central_noise_indices] = float('NaN')
        data[:,sf.float_to_int(hdr['NAXIS2']/2.-buffer):sf.float_to_int(hdr['NAXIS2']/2.+buffer),\
                        sf.float_to_int(hdr['NAXIS1']/2.-buffer):sf.float_to_int(hdr['NAXIS1']/2.+buffer)] =\
                        central
        print_log(f'''PREPROCESSING: Your cube had a significant amount of values below -10*sigma. If you do not have a central absorption source there is something seriously wrong with the cube.
{"":8s}PREPROCESSING: We blanked these values.
''',Configuration)


    # Check whether any channels got fully blanked
    blanked_channels = []
    for z in range(hdr['NAXIS3']):
        if np.isnan(data[z,:,:]).all():
            blanked_channels.append(str(z))
    if len(blanked_channels) == 0.:
        blanked_channels = ['-1']
    hdr['BL_CHAN'] = ','.join(blanked_channels)
    #Previously SoFiA was not dealing with blanks properly and the statement went here
    #finally we want to get the noise level from the negative in the final cube

    new_noise = np.nanstd(np.hstack([data[data<0],-1.*data[data<0]]))
    #If we have noiseles cubes this will be far to low
    if abs(new_noise/np.mean([channel_noise,noise_corner])) > 4. \
        or abs(new_noise/np.mean([channel_noise,noise_corner])) < 0.25:
            print_log(f'''PREPROCESSING: There is something odd in the noise statistics of your cube. We are not using the nehgative values
''',Configuration)
            #new_noise = np.mean([channel_noise,noise_corner])
    hdr['FATNOISE'] = new_noise

    return data
prep_cube.__doc__ =f'''
 NAME:
    prep_cube

 PURPOSE:
    Clean up the cube  and make sure it has all the right
    characteristics and blanks

 CATEGORY:
    fits_functions

 INPUTS:
    Configuration = Standard FAT configuration
    hdr = header to be cleaned
    data = data array to be cleaned

 OPTIONAL INPUTS:

 OUTPUTS:
    the updated header

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''

#Create an optimized cube if required
def optimized_cube(Configuration,Fits_Files):
    pix_per_beam = Configuration['BEAM_IN_PIXELS'][1]
    if pix_per_beam > Configuration['OPT_PIXEL_BEAM']:
        cube = fits.open(Configuration['FITTING_DIR']+Fits_Files['FITTING_CUBE'])
        data = cube[0].data
        hdr = cube[0].header
        if f"{abs(hdr['CDELT1']):.16f}" != f"{abs(hdr['CDELT2']):.16f}":
            print_log(f'''OPTIMIZED_CUBE: Your input cube does not have square pixels.
{"":8s}OPTIMIZED_CUBE: FAT cannot optimize your cube.
''', Configuration)
        cube.close()
        required_cdelt = hdr['BMIN']/Configuration['OPT_PIXEL_BEAM']
        ratio = required_cdelt/abs(hdr['CDELT2'])
        opt_data,opt_hdr = regrid_cube(data, hdr, ratio)

        fits.writeto(Configuration['FITTING_DIR']+Fits_Files['OPTIMIZED_CUBE'], opt_data,opt_hdr)

        Configuration['OPTIMIZED'] = True
        print_log(f'''OPTIMIZED_CUBE: Your input cube has {pix_per_beam} pixels along the minor FWHM.
{"":8s}OPTIMIZED_CUBE: You requested { Configuration['OPT_PIXEL_BEAM']} therefore we regridded the cube into a new cube.
{"":8s}OPTIMIZED_CUBE: We are using the pixel size of {required_cdelt}.
''', Configuration)

    else:
        print_log(f'''OPTIMIZED_CUBE: Your input cube has {pix_per_beam} pixels along the minor FWHM.
{"":8s}OPTIMIZED_CUBE: You requested {Configuration['OPT_PIXEL_BEAM']} but we cannot improve the resolution.
{"":8s}OPTIMIZED_CUBE: We are using the pixel size of the original cube.
''', Configuration)
optimized_cube.__doc__ =f'''
 NAME:
    optimized_cube

 PURPOSE:
    Check the amount of pixels that are in the minor axis FWHM,
    if more than OPT_PIXELBEAM regrid the cube to less pixels
    and write FAT_opt cube and set Configuration['OPTIMIZED'] = True

 CATEGORY:
    fits_functions

 INPUTS:
    Configuration = Standard FAT configuration
    Fits_Files = Standard FAT dictionary with filenames

 OPTIONAL INPUTS:

 OUTPUTS:
    An optimized cube is created and written to disk.

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    Unspecified

 NOTE:
'''

def regrid_cube(data,hdr,ratio):
    regrid_hdr = copy.deepcopy(hdr)
    # First get the shape of the data
    shape = np.array(data.shape, dtype=float)
    #The new shape has to be integers
    new_shape = [sf.float_to_int(x/ratio) for x in shape]
    # Which means our real ratio is
    real_ratio = shape/new_shape
    #* np.ceil(shape / ratio).astype(int)
    new_shape[0] = shape[0]
    # Create the zero-padded array and assign it with the old density
    regrid_data = regridder(data,new_shape)
    regrid_hdr['CRPIX1'] =   (regrid_hdr['CRPIX1']-1)/real_ratio[2]+1
    regrid_hdr['CRPIX2'] =   (regrid_hdr['CRPIX2']-1)/real_ratio[1]+1
    wcs_found = False
    try:
        regrid_hdr['CDELT1'] =   regrid_hdr['CDELT1']*real_ratio[2]
        regrid_hdr['CDELT2'] =   regrid_hdr['CDELT2']*real_ratio[1]
        wcs_found = True
    except KeyError:
        print("No CDELT found")
    try:
        regrid_hdr['CD1_1'] =   regrid_hdr['CD1_1']*real_ratio[2]
        regrid_hdr['CD2_2'] =   regrid_hdr['CD2_2']*real_ratio[1]
        wcs_found = True
    except KeyError:
        if not wcs_found:
            print("No CD corr matrix found")
    try:
        regrid_hdr['CD1_2'] =   regrid_hdr['CD1_2']*real_ratio[2]
        regrid_hdr['CD2_1'] =   regrid_hdr['CD2_1']*real_ratio[1]
        wcs_found = True
    except KeyError:
        if not wcs_found:
            print("No CD cross-corr matrix found")

    regrid_hdr['NAXIS1'] =   new_shape[2]
    regrid_hdr['NAXIS2'] =   new_shape[1]
    return regrid_data,regrid_hdr
regrid_cube.__doc__ =f'''
 NAME:
    regrid_cube

 PURPOSE:
    Regrid a cube to an arbitrary smaller cube in the spatial plane with larger pixels

 CATEGORY:
    fits_functions

 INPUTS:
    data = the array containing the input cube.
    hdr = the header data
    ratio = the requested ratio of old/new

 OPTIONAL INPUTS:

 OUTPUTS:
    regriddata = regridded data array
    regrid_hdr = header corresponding to the regridded data

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    copy.deepcopy, regridder, np.array

 NOTE: This might not work well with extended projections.
'''

def regridder(oldarray, newshape):
    oldshape = np.array(oldarray.shape)
    newshape = np.array(newshape, dtype=float)
    ratios = oldshape/newshape
        # calculate new dims
    nslices = [ slice(0,j) for j in list(newshape) ]
    #make a list with new coord
    new_coordinates = np.mgrid[nslices]
    #scale the new coordinates
    for i in range(len(ratios)):
        new_coordinates[i] *= ratios[i]
    #create our regridded array
    newarray = ndimage.map_coordinates(oldarray, new_coordinates,order=1)

    return newarray
regridder.__doc__ =f'''
 NAME:
regridder
 PURPOSE:
Regrid an array into a new shape through the ndimage module
 CATEGORY:
    fits_functions

 INPUTS:
    oldarray = the larger array
    newshape = the new shape that is requested

 OPTIONAL INPUTS:

 OUTPUTS:
    newarray = regridded array

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:
    scipy.ndimage.map_coordinates, np.array, np.mgrid

 NOTE:
'''

def smooth_mask(Configuration,mask):
    print_log(f'''SMOOTH_MASK: Starting smooth mask
''', Configuration)

    # First we smooth the mask
    mask_data = np.array(mask[0].data,dtype=float)
    mask_data = ndimage.gaussian_filter(mask_data,sigma=(0.,\
        Configuration['BEAM_IN_PIXELS'][1] / np.sqrt(8 * np.log(2)),\
        Configuration['BEAM_IN_PIXELS'][0] / np.sqrt(8 * np.log(2))),order=0)
    mask_data[mask_data > 0.05] = 1.
    mask_data[mask_data < 0.05] = 0.
    mask[0].data=mask_data
    return mask
    #reaplly it to get the moments


    # If we have a lot of pixels and stuff we need smooth more
    #So we take the inverse of the initial factor with a standard of 0.75 minimum of 0.5 and a maximum of 1.25

smooth_mask.__doc__ =f'''
 NAME:
    smooth_mask
 PURPOSE:
    smooth the mask with a beam
 CATEGORY:
    fits_functions

 INPUTS:

 OPTIONAL INPUTS:

 OUTPUTS:

 OPTIONAL OUTPUTS:

 PROCEDURES CALLED:

 NOTE:
'''
